from dataclasses import dataclass


@dataclass
class Rect:
    """Data structure to represent and transform rectangles.

    Note: Using OpenCV style geometry ordering (y, x, height, width)
    """

    top: int
    left: int
    width: int
    height: int

    def __str__(self) -> str:
        return f"Rect(x={self.left}, y={self.top}, w={self.width}, h={self.height})"

    @property
    def right(self) -> int:
        return self.left + self.width

    @property
    def bottom(self) -> int:
        return self.top + self.height

    @property
    def left_top(self) -> tuple[int, int]:
        return (self.left, self.top)

    @property
    def right_bottom(self) -> tuple[int, int]:
        return (self.left + self.width, self.top + self.height)

    @property
    def area(self) -> int:
        return abs(self.width * self.height)

    @property
    def geometry(self) -> tuple[int, int, int, int]:
        return (self.top, self.left, self.height, self.width)

    def pad(self, padding: int) -> None:
        """Centered padding of the rectangle.

        Args:
            padding: Absolute padding value.
        """
        self.top = self.top - padding
        self.left = self.left - padding
        self.width = self.width + 2 * padding
        self.height = self.height + 2 * padding

    def scale(self, factor: float) -> None:
        """Centered scaling of the rectangle.

        Note: Using int() instead of more accurate round(), because we care less about
          single pixel accuracy than about speed.

        Args:
            factor: Multiplication factor.
        """
        new_width = int(self.width * factor)
        new_height = int(self.height * factor)
        self.top -= int((new_height - self.height) / 2)
        self.left -= int((new_width - self.width) / 2)
        self.width = new_width
        self.height = new_height

    def copy(self) -> "Rect":
        return Rect(top=self.top, left=self.left, height=self.height, width=self.width)

    def move_by(self, y: int, x: int) -> None:
        """Move the rectangle by the provided x and y values.

        Args:
            y: Value to move the rectangle by on the y-axis.
            x: Value to move the rectangle by on the x-axis.
        """
        self.left += x
        self.top += y

    def stay_within(self, height: int, width: int) -> None:
        """Ensure the rectangle is within the bounds of the provided width and height.

        Preserves aspect ratio.

        Args:
            height: Height of the bounding box.
            width: Width of the bounding box.
        """
        # Scale self to fit within width and height
        scale_factor = min(width / self.width, height / self.height)
        if scale_factor < 1:
            self.scale(scale_factor)

        self.top = min(max(0, self.top), height - self.height)
        self.left = min(max(0, self.left), width - self.width)



import logging
import time
from pathlib import Path
from threading import Thread
from typing import Any

import cv2
import numpy as np

logger = logging.getLogger(__name__)


class DemoVideoCapture:
    def __init__(self) -> None:
        self._demo_video_file = str(Path(__file__).parent / "resources" / "demo.mp4")
        self.capture = cv2.VideoCapture(self._demo_video_file)
        self.fps = self.capture.get(cv2.CAP_PROP_FPS)
        self.last_frame = time.perf_counter()

    def read(self) -> tuple[bool, np.ndarray]:
        while self.last_frame + 1 / self.fps > time.perf_counter():
            pass
        self.last_frame = time.perf_counter()
        return_code, frame = self.capture.read()
        if not return_code:
            self.capture.set(cv2.CAP_PROP_POS_FRAMES, 0)
            return_code, frame = self.capture.read()
        return (return_code, frame)

    def release(self) -> None:
        pass

    def isOpened(self) -> bool:  # noqa: N802 # camelCase used by OpenCV
        return True

    def set(self, _: Any, __: Any) -> int:  # noqa: ANN401
        return 0


class FallbackVideoCapture:
    def __init__(self) -> None:
        self._image_path = str(Path(__file__).parent / "resources" / "fallback.png")
        self.frame = cv2.imread(self._image_path)

    def read(self) -> tuple[bool, np.ndarray]:
        # Add some noise to invalidate cache
        return (
            True,
            self.frame - np.random.randint(0, 3, size=self.frame.shape, dtype=np.uint8),
        )

    def release(self) -> None:
        pass

    def isOpened(self) -> bool:  # noqa: N802 # camelCase used by OpenCV
        return True

    def set(self, _: Any, __: Any) -> int:  # noqa: ANN401
        return 0


class Camera:
    def __init__(self) -> None:
        self.DEMO_CAM_ID = 98
        self.FALLBACK_CAM_ID = 99
        self.available_cameras = self._get_available_cameras()
        self.cam_id: int
        self._capture: (
            cv2.VideoCapture | DemoVideoCapture | FallbackVideoCapture | None
        ) = None
        self.frame: np.ndarray = np.zeros((1080, 1920, 3), np.uint8)
        self.fps: list[float] = [0]
        self.fps_window = 100
        self.stop_video_thread = False
        self.video_thread: Thread | None = None

    def _get_video_capture(
        self, cam_id: int
    ) -> cv2.VideoCapture | DemoVideoCapture | FallbackVideoCapture:
        if cam_id == self.DEMO_CAM_ID:
            logger.info("Using demo video capture.")
            return DemoVideoCapture()

        if cam_id == self.FALLBACK_CAM_ID:
            logger.info("Using fallback video capture.")
            return FallbackVideoCapture()

        return cv2.VideoCapture(cam_id, cv2.CAP_V4L2)

    def _get_available_cameras(self) -> dict[int, np.ndarray]:
        """Heuristically determine available video inputs.

        OpenCV does not provide a reliable way to list available cameras. Therefore,
        we try to open /video0 - /video9 and check if we can read some metadata. If we
        can't we deduce that the camera is not available, but it might just be already
        in use. We also won't detect inputs with a higher index.

        Returns:
            IDs of available cameras
        """
        cams = {}
        cam_ids_to_try = [*range(10), self.DEMO_CAM_ID, self.FALLBACK_CAM_ID]

        for idx in cam_ids_to_try:
            cap = self._get_video_capture(idx)
            cap.set(cv2.CAP_PROP_FRAME_WIDTH, 1920)  # type: ignore # FP
            cap.set(cv2.CAP_PROP_FRAME_HEIGHT, 1080)  # type: ignore # FP
            try:
                read_status, frame = cap.read()
            except cv2.error:
                logger.debug("Camera at /video%s seems unavailable (cv2.error)", idx)
            else:
                if read_status:
                    logger.debug("Camera at /video%s is available", idx)
                    cams[idx] = frame
                else:
                    logger.debug("Camera at /video%s seems unavailable (no frame)", idx)
            finally:
                cap.release()

        logger.info("Available cameras: %s", cams.keys())
        return cams

    def start(self, cam_id: int) -> None:
        if self.video_thread is not None:
            raise ValueError("Camera needs to be stopped before starting!")

        first_cam_id = next(iter(self.available_cameras), None)
        cam_is_available = cam_id in self.available_cameras

        if cam_is_available:
            logger.info("Using camera %s.", cam_id)
            self.cam_id = cam_id
        elif not cam_is_available and first_cam_id is not None:
            logger.warning("Camera %s not available. Fallback to first one.", cam_id)
            self.cam_id = first_cam_id
        else:
            logger.error("No camera accessible! Is another application using it?")
            self.cam_id = 99

        self._capture = self._get_video_capture(cam_id=self.cam_id)

        # Set compressed codec for way better performance:
        self._capture.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*"MJPG"))  # type: ignore # FP

        # Max resolution & FPS. OpenCV automatically selects lower one, if needed:
        self._capture.set(cv2.CAP_PROP_FPS, 60)  # type: ignore # FP
        self._capture.set(cv2.CAP_PROP_FRAME_WIDTH, 1920)  # type: ignore # FP
        self._capture.set(cv2.CAP_PROP_FRAME_HEIGHT, 1080)  # type: ignore # FP

        self.stop_video_thread = False
        self.video_thread = Thread(target=self.update, args=())
        self.video_thread.start()

    def stop(self) -> None:
        if self.video_thread is None:
            return

        self.stop_video_thread = True
        self.video_thread.join()

        if self._capture and self._capture.isOpened():
            self._capture.release()

        self.video_thread = None

    def get_frame(self) -> np.ndarray:
        return self.frame

    def update(self) -> None:
        logger.info("Camera thread started.")
        clock_period = 1 / cv2.getTickFrequency()
        last_tick = cv2.getTickCount()
        while not self.stop_video_thread:
            try:
                if not self._capture:
                    logger.error("Capture device not ready.")
                    break

                _, self.frame = self._capture.read()

                tick = cv2.getTickCount()
                fps = 1 / ((tick - last_tick) * clock_period)
                last_tick = tick

                self.fps.append(fps)
                if len(self.fps) > self.fps_window:
                    self.fps.pop(0)

            except cv2.error:  # type: ignore # FP
                logger.exception("Error in camera update.")
                break

        logger.info("Camera thread stopped.")



import tempfile

import cv2
import gi
import numpy as np

gi.require_version("Gtk", "4.0")
from gi.repository import Gtk  # noqa: E402


def cv2_image_to_gtk_image(cv2_image: np.ndarray) -> Gtk.Image:
    """Create Gtk.Image from cv2 image via a temporary file.

    Args:
        cv2_image: OpenCV input image.

    Returns:
        GTK image widget.
    """
    with tempfile.NamedTemporaryFile(suffix=".jpg") as temp_file:
        temp_image = cv2.resize(cv2_image, fx=0.20, fy=0.20, dsize=(0, 0))
        cv2.imwrite(temp_file.name, temp_image)
        image = Gtk.Image.new_from_file(temp_file.name)
    return image



import logging
from pathlib import Path

import cv2
import numpy as np

from myhumbleself.structures import Rect

logger = logging.getLogger(__name__)


class FaceDetection:
    cnn_onnx = str(
        Path(__file__).parent / "resources" / "face_detection_yunet_2023mar_int8.onnx"
    )

    def __init__(self) -> None:
        self._history: list[Rect] = []
        self._max_history_len = 20
        self._last_smoothed_geometry: Rect | None = None
        self._detector_cnn = cv2.FaceDetectorYN.create(self.cnn_onnx, "", (42, 42))

        self.fluctuation_threshold_factor = 0.03
        self.follow_face_speed_factor = 0.2
        self.debug_mode = False

    @staticmethod
    def _draw_bounding_box(
        image: np.ndarray, face_coords: Rect, color: tuple[int, int, int]
    ) -> None:
        cv2.rectangle(
            img=image,
            pt1=face_coords.left_top,
            pt2=face_coords.right_bottom,
            color=color,
            thickness=2,
        )

    def _detect_faces_cnn(self, image: np.ndarray) -> list[Rect]:
        # Scale down to speed up and improve detection
        # TODO: Make instance variable, expose via cli?
        target_width = 250  # based on little testing
        scale_factor = target_width / max(image.shape)
        image = cv2.resize(
            image,
            None,
            fx=scale_factor,
            fy=scale_factor,
            interpolation=cv2.INTER_NEAREST,
        )

        # Detect faces
        self._detector_cnn.setInputSize((image.shape[1], image.shape[0]))
        face_detections = self._detector_cnn.detect(image)

        # Convert to Rect objects and scale back up
        faces: list[Rect] = []
        if face_detections[1] is not None:
            for data in face_detections[1]:
                left = int(data[0] / scale_factor)
                top = int(data[1] / scale_factor)
                width = int(data[2] / scale_factor)
                height = int(data[3] / scale_factor)
                faces.append(Rect(left=left, top=top, width=width, height=height))

        return faces

    @staticmethod
    def _select_largest_face(faces: list[Rect]) -> Rect | None:
        largest_face = None

        for face in faces:
            if largest_face is None or face.area > largest_face.area:
                largest_face = face

        return largest_face

    def _smooth_geometry(self) -> Rect:
        mean = np.mean([f.geometry for f in self._history], axis=0, dtype=int).tolist()
        smoothed_geometry = Rect(
            top=mean[0], left=mean[1], height=mean[2], width=mean[3]
        )

        if not self._last_smoothed_geometry:
            # No last value, nothing to smooth
            self._last_smoothed_geometry = smoothed_geometry
            return smoothed_geometry

        # Stabilize dimensions by comparing with the last mean_rect
        for attr in ["left", "top", "width", "height"]:
            new_val = getattr(smoothed_geometry, attr)
            old_val = getattr(self._last_smoothed_geometry, attr)
            if abs(new_val - old_val) <= old_val * self.fluctuation_threshold_factor:
                # Use old value to avoid fluctuations
                setattr(smoothed_geometry, attr, old_val)
            else:
                # Move into direction of the new value, but smoothly to avoid jumps
                distance = new_val - old_val
                step_size = abs(distance) * self.follow_face_speed_factor
                correction_step = int(np.sign(distance) * max(1, step_size))
                setattr(smoothed_geometry, attr, old_val + correction_step)

        self._last_smoothed_geometry = smoothed_geometry
        return smoothed_geometry

    def get_face(self, image: np.ndarray) -> Rect:
        if not self._history:
            # Start with full image
            self._history.append(
                Rect(top=0, left=0, width=image.shape[1] - 1, height=image.shape[0] - 1)
            )

        faces = self._detect_faces_cnn(image)

        if self.debug_mode:
            for f in faces:
                self._draw_bounding_box(image, f, color=(0, 125, 0))

        face = self._select_largest_face(faces=faces)

        if face:
            self._history.append(face)
            self._history = self._history[-self._max_history_len :]

        face = self._smooth_geometry()
        return face



import pytest

from myhumbleself import structures


def test_rect_repr():
    y, x, h, w = (5, 10, 20, 40)
    rect = structures.Rect(top=y, left=x, height=h, width=w)
    rect_repr = str(rect)
    assert f"y={y}" in rect_repr
    assert f"x={x}" in rect_repr
    assert f"w={w}" in rect_repr
    assert f"h={h}" in rect_repr


@pytest.mark.parametrize(
    ("left", "width", "expected_right"),
    [
        (0, 100, 100),
        (100, 100, 200),
        (-100, 100, 0),
        (-200, 100, -100),
        (50, -100, -50),
    ],
)
def test_rect_right(left, width, expected_right):
    rect = structures.Rect(top=0, left=left, height=100, width=width)
    assert rect.right == expected_right


@pytest.mark.parametrize(
    ("top", "height", "expected_bottom"),
    [
        (0, 100, 100),
        (100, 100, 200),
        (-100, 100, 0),
        (-200, 100, -100),
        (50, -100, -50),
    ],
)
def test_rect_bottom(top, height, expected_bottom):
    rect = structures.Rect(top=top, left=0, height=height, width=100)
    assert rect.bottom == expected_bottom


def test_rect_left_top():
    left, top = 10, 20
    rect = structures.Rect(top=top, left=left, width=10, height=10)
    assert rect.left_top == (left, top)


def test_rect_right_bottom():
    right, bottom = 10, 20
    rect = structures.Rect(top=0, left=0, width=right, height=bottom)
    assert rect.right_bottom == (right, bottom)


@pytest.mark.parametrize(
    ("yxhw", "expected_area"),
    [
        ((0, 0, 50, 100), 5000),
        ((-50, -100, 50, 100), 5000),
        ((0, 0, -50, 100), 5000),
    ],
)
def test_rect_area(yxhw, expected_area):
    rect = structures.Rect(top=yxhw[0], left=yxhw[1], height=yxhw[2], width=yxhw[3])
    assert rect.area == expected_area


def test_rect_geometry():
    y, x, h, w = (5, 10, 20, 40)
    rect = structures.Rect(top=y, left=x, height=h, width=w)
    assert rect.geometry == (y, x, h, w)


@pytest.mark.parametrize(
    ("yxhw", "factor", "expected_yxhw"),
    [
        ((0, 0, 50, 100), 2.0, (-25, -50, 100, 200)),
        ((-100, -100, 50, 100), 2.0, (-125, -150, 100, 200)),
        ((-10, -10, 5, 20), 0.5, (-9, -5, 2, 10)),
    ],
)
def test_rect_scale(yxhw, factor, expected_yxhw):
    rect = structures.Rect(top=yxhw[0], left=yxhw[1], height=yxhw[2], width=yxhw[3])
    rect.scale(factor=factor)
    assert rect.top == expected_yxhw[0]
    assert rect.left == expected_yxhw[1]
    assert rect.height == expected_yxhw[2]
    assert rect.width == expected_yxhw[3]


@pytest.mark.parametrize(
    ("yxhw", "padding", "expected_yxhw"),
    [
        ((0, 0, 5, 10), 5, (-5, -5, 15, 20)),
        ((-10, -5, 5, 10), 2, (-12, -7, 9, 14)),
        ((10, 20, 30, 40), -2, (12, 22, 26, 36)),
    ],
)
def test_rect_pad(yxhw, padding, expected_yxhw):
    rect = structures.Rect(top=yxhw[0], left=yxhw[1], height=yxhw[2], width=yxhw[3])
    rect.pad(padding)
    assert rect.top == expected_yxhw[0]
    assert rect.left == expected_yxhw[1]
    assert rect.height == expected_yxhw[2]
    assert rect.width == expected_yxhw[3]


def test_rect_copy():
    rect = structures.Rect(0, 5, 10, 15)
    rect_copy = rect.copy()
    assert rect.top == rect_copy.top
    assert rect.left == rect_copy.left
    assert rect.height == rect_copy.height
    assert rect.width == rect_copy.width
    assert rect is not rect_copy


def test_rect_move_by():
    y, x, h, w = (5, 10, 20, 40)
    delta_y, delta_x = -10, 5
    rect = structures.Rect(top=y, left=x, height=h, width=w)
    rect.move_by(y=delta_y, x=delta_x)
    assert rect.top == y + delta_y
    assert rect.left == x + delta_x
    assert rect.height == h
    assert rect.width == w


@pytest.mark.parametrize(
    ("yxhw", "max_height", "max_width", "expected_yxhw"),
    [
        ((5, 5, 10, 10), 20, 20, (5, 5, 10, 10)),
        ((0, 0, 20, 20), 10, 10, (0, 0, 10, 10)),
        ((0, 5, 20, 20), 10, 10, (0, 0, 10, 10)),
        ((0, 5, 20, 10), 10, 10, (0, 5, 10, 5)),
        ((10, 20, 40, 60), 20, 10, (14, 0, 6, 10)),
    ],
)
def test_rect_stay_within(yxhw, max_height, max_width, expected_yxhw):
    rect = structures.Rect(top=yxhw[0], left=yxhw[1], height=yxhw[2], width=yxhw[3])
    rect.stay_within(height=max_height, width=max_width)
    assert rect.top == expected_yxhw[0]
    assert rect.left == expected_yxhw[1]
    assert rect.height == expected_yxhw[2]
    assert rect.width == expected_yxhw[3]



import argparse
import logging
import os
import platform
import time
from pathlib import Path
from statistics import mean

# Hide warnings shown during search for cameras
os.environ["OPENCV_LOG_LEVEL"] = "FATAL"

import gi

from myhumbleself import __version__, config, converters, video_handler

gi.require_version("Gdk", "4.0")
gi.require_version("Gtk", "4.0")
gi.require_version("Gio", "2.0")
gi.require_version("GdkPixbuf", "2.0")
from gi.repository import Gdk, GdkPixbuf, Gio, Gtk  # noqa: E402

logger = logging.getLogger(__name__)


def init_logger(log_level: str = "WARNING") -> None:
    log_format = "%(asctime)s - %(levelname)-7s - %(name)s:%(lineno)d - %(message)s"
    datefmt = "%H:%M:%S"
    logging.basicConfig(format=log_format, datefmt=datefmt, level=log_level)


class MyHumbleSelf(Gtk.Application):
    def __init__(self, application_id: str, args: argparse.Namespace) -> None:
        super().__init__(application_id=application_id)

        # Resource file
        self.resource = Gio.resource_load(
            str(Path(__file__).parent / "resources" / "myhumbleself.gresource")
        )
        Gio.Resource._register(self.resource)

        # Top level
        self.win: Gtk.ApplicationWindow

        # Webcam widget
        self.picture: Gtk.Picture

        # Headerbar widgets
        self.follow_face_button: Gtk.ToggleButton
        self.shape_box: Gtk.FlowBox
        self.camera_box: Gtk.FlowBox

        # Controls Container
        self.overlay: Gtk.Overlay

        # Controls
        self.reset_button: Gtk.Button
        self.toggle_controls_button: Gtk.Button
        self.right_button: Gtk.Button
        self.left_button: Gtk.Button
        self.down_button: Gtk.Button
        self.up_button: Gtk.Button
        self.zoom_in_button: Gtk.Button
        self.zoom_out_button: Gtk.Button

        # Init values
        self.config = config.load()
        self.fps: list[float] = [0]
        self.fps_window = 50
        self.cam_item_prefix = "/dev/video"
        self.loglevel_debug = logger.getEffectiveLevel() == logging.DEBUG
        self.last_image_id = b""
        self.video_handler = video_handler.VideoHandler(
            cam_id=self.config["main"].getint("last_active_camera", 0),
            shape_png_buffer=self._load_active_shape_png(),
            zoom_factor=self.config["main"].getfloat("zoom_factor", 1),
            offset_x=self.config["main"].getint("offset_x", 0),
            offset_y=self.config["main"].getint("offset_y", 0),
            follow_face=self.config["main"].getboolean("follow_face", True),
        )

        self.connect("activate", self.on_activate)
        self.connect("shutdown", self.on_shutdown)

    def on_activate(self, app: Gtk.Application) -> None:
        """Initialize window on application activation.

        Args:
            app: Gtk Application.
        """
        self.builder = Gtk.Builder()
        self.builder.add_from_resource("/com/github/dynobo/myhumbleself/window.ui")

        theme = Gtk.IconTheme.get_for_display(Gdk.Display.get_default())
        theme.add_resource_path("/com/github/dynobo/myhumbleself/icons")

        self.win = self.builder.get_object("main_window")
        self.win.set_application(self)
        self.win.set_icon_name("com.github.dynobo.myhumbleself")

        picture = self.builder.get_object("picture")
        picture.add_tick_callback(self.on_picture_tick)

        self.shape_box = self.init_shape_box()
        self.follow_face_button = self.init_follow_face_button()
        self.camera_box = self.init_camera_box()
        self.about_button = self.builder.get_object("about_button")
        self.about_button.connect("clicked", lambda _: self.show_about_dialog())
        self.debug_mode_button = self.builder.get_object("debug_mode_button")
        if self.loglevel_debug:
            self.debug_mode_button.set_visible(True)
            self.debug_mode_button.connect("clicked", self.on_toggle_debug_position)

        self.overlay = self.builder.get_object("overlay")

        self.toggle_controls_button = self.builder.get_object("toggle_controls_button")
        self.toggle_controls_button.set_icon_name("controls-hide-symbolic")
        self.toggle_controls_button.connect("toggled", self.on_toggle_controls_clicked)

        self.reset_button = self.builder.get_object("reset_button")
        self.reset_button.set_icon_name("reset-symbolic")
        self.reset_button.set_tooltip_text("Reset view")
        self.reset_button.connect("clicked", self.on_reset_clicked)

        self.zoom_in_button = self.builder.get_object("zoom_in_button")
        self.zoom_in_button.connect("clicked", self.on_zoom, 1)

        self.zoom_out_button = self.builder.get_object("zoom_out_button")
        self.zoom_out_button.connect("clicked", self.on_zoom, -1)

        self.left_button = self.builder.get_object("left_button")
        self.left_button.connect("clicked", self.on_move_clicked, -1, 0)

        self.right_button = self.builder.get_object("right_button")
        self.right_button.connect("clicked", self.on_move_clicked, 1, 0)

        self.up_button = self.builder.get_object("up_button")
        self.up_button.connect("clicked", self.on_move_clicked, 0, -1)

        self.down_button = self.builder.get_object("down_button")
        self.down_button.connect("clicked", self.on_move_clicked, 0, 1)

        self.init_css()

        self.win.present()

    def show_about_dialog(self) -> None:
        self.about = Gtk.AboutDialog()
        self.about.set_transient_for(self.win)
        self.about.set_modal(True)
        self.about.set_program_name("MyHumbleSelf")
        self.about.set_logo_icon_name("com.github.dynobo.myhumbleself")
        self.about.set_license_type(Gtk.License.GPL_3_0)
        self.about.set_comments(
            "Utility to display webcam image for presentations or screencasts on Linux."
        )
        self.about.set_website("https://github.com/dynobo/myhumbleself")
        self.about.set_website_label("Github")
        self.about.set_version(__version__)
        self.about.set_system_information(self.get_system_info())
        self.about.set_visible(True)

    def create_camera_menu_button(self, cam_id: int) -> Gtk.ToggleButton:
        """Create a custom button for camera menu, with image and label underneath.

        Args:
            cam_id: ID of the camera for which the button is created.

        Returns:
            Button widget.
        """
        image = converters.cv2_image_to_gtk_image(
            self.video_handler.available_cameras[cam_id]
        )
        label = Gtk.Label()
        label.set_text(f"{self.cam_item_prefix}{cam_id}")

        button_box = Gtk.Box()
        button_box.set_orientation(Gtk.Orientation.VERTICAL)
        button_box.append(image)
        button_box.append(label)

        button = Gtk.ToggleButton()
        button.set_size_request(56, 56)
        button.set_has_frame(False)
        button.set_child(button_box)
        button.connect("toggled", self.on_camera_toggled, cam_id)
        button.set_css_classes([*button.get_css_classes(), "camera-button"])
        return button

    def init_camera_box(self) -> Gtk.FlowBox:
        """Fill the camera menu's flow box with buttons for each camera.

        Also hide the camera menu if only one camera is available.

        Returns:
            Widget containing the camera selection buttons.
        """
        camera_menu_button = self.builder.get_object("camera_menu_button")
        camera_box = self.builder.get_object("camera_box")
        first_button = None
        for cam_id in self.video_handler.available_cameras:
            button = self.create_camera_menu_button(cam_id)

            # Activate button if it was the last active camera
            if cam_id == self.config["main"].getint("last_active_camera", 0):
                button.set_active(True)

            if cam_id in [
                self.video_handler.FALLBACK_CAM_ID,
                self.video_handler.DEMO_CAM_ID,
            ]:
                button.set_visible(self.loglevel_debug)

            # Set button group
            if first_button is None:
                first_button = button
            else:
                button.set_group(first_button)

            camera_box.append(button)

        # Hide camera menu if only one camera (plus fallback) is available, except
        # when in debug mode:
        is_visible = (
            len(self.video_handler.available_cameras) - 1 > 1 or self.loglevel_debug
        )
        camera_menu_button.set_visible(is_visible)

        return camera_box

    def init_shape_box(self) -> Gtk.FlowBox:
        """Setup widget for selecting shape overlay.

        Returns:
            Widget containing shape selection buttons.
        """
        shape_menu_button = self.builder.get_object("shape_menu_button")
        shape_menu_button.set_icon_name("shapes-symbolic")

        shape_box = self.builder.get_object("shape_box")

        first_button = None
        for shape in self.resource.enumerate_children(
            "/com/github/dynobo/myhumbleself/shapes", Gio.ResourceLookupFlags.NONE
        ):
            button = Gtk.ToggleButton()
            button.set_size_request(56, 56)
            button.set_icon_name(f"{shape[:-4]}-symbolic")
            button.set_has_frame(False)
            button.connect("toggled", self.on_shape_toggled, shape)
            button.set_css_classes([*button.get_css_classes(), "shape-button"])

            # Activate stored shape:
            if shape == self.config["main"].get("shape"):
                button.set_active(True)

            # Set button group
            if first_button is None:
                first_button = button
            else:
                button.set_group(first_button)

            shape_box.append(button)

        return shape_box

    def init_follow_face_button(self) -> Gtk.ToggleButton:
        """Setup widget for toggling face detection mode.

        Returns:
            Toggle Button.
        """
        is_follow_face = self.config["main"].getboolean("follow_face")
        follow_face_button = self.builder.get_object("follow_face_button")
        follow_face_button.set_active(is_follow_face)
        follow_face_button.set_tooltip_text(
            "Do not follow face" if is_follow_face else "Follow face"
        )
        follow_face_button.set_icon_name(
            "follow-face-off-symbolic" if is_follow_face else "follow-face-symbolic"
        )
        follow_face_button.connect("clicked", self.on_follow_face_clicked)
        return follow_face_button

    def init_css(self) -> None:
        """Apply style from css file to the application."""
        self.css_provider = self.builder.get_object("css_provider")
        Gtk.StyleContext.add_provider_for_display(
            Gdk.Display.get_default(),
            self.css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )
        self.css_provider.load_from_resource(
            "/com/github/dynobo/myhumbleself/style.css"
        )

    def on_follow_face_clicked(self, button: Gtk.ToggleButton) -> None:
        follow_face = button.get_active()
        self.config.set_persistent("follow_face", follow_face)
        self.video_handler.follow_face = follow_face
        if follow_face:
            button.set_tooltip_text("Do not follow face")
            button.set_icon_name("follow-face-off-symbolic")
        else:
            button.set_tooltip_text("Follow face")
            button.set_icon_name("follow-face-symbolic")

    def _load_active_shape_png(self) -> bytes:
        shape = self.config["main"].get("shape")
        shape_png = self.resource.lookup_data(
            f"/com/github/dynobo/myhumbleself/shapes/{shape}",
            Gio.ResourceLookupFlags.NONE,
        ).get_data()
        return shape_png

    def on_shape_toggled(self, button: Gtk.ToggleButton, shape: str) -> None:
        if not button.get_active():
            return
        self.config.set_persistent("shape", shape)
        self.video_handler.set_shape(self._load_active_shape_png())

    def on_camera_toggled(self, button: Gtk.ToggleButton, cam_id: int) -> None:
        if not button.get_active():
            return
        self.video_handler.set_camera(cam_id)
        self.config.set_persistent("last_active_camera", cam_id)

    def on_reset_clicked(self, button: Gtk.Button) -> None:
        self.config.set_persistent("offset_x", 0)
        self.config.set_persistent("offset_y", 0)
        self.config.set_persistent("zoom_factor", 1)
        self.video_handler.reset_view()

    def on_move_clicked(self, button: Gtk.Button, factor_x: int, factor_y: int) -> None:
        self.video_handler.offset_x += factor_x * self.video_handler.MOVE_STEP
        self.video_handler.offset_y += factor_y * self.video_handler.MOVE_STEP
        self.config.set_persistent("offset_x", self.video_handler.offset_x)
        self.config.set_persistent("offset_y", self.video_handler.offset_y)

    def on_zoom(self, _: Gtk.Button, factor_z: int) -> None:
        zoom = self.config["main"].getfloat("zoom_factor", 1)
        zoom -= self.video_handler.ZOOM_STEP * factor_z
        self.video_handler.zoom_factor = max(zoom, self.video_handler.MIN_ZOOM_FACTOR)
        self.config.set_persistent("zoom_factor", zoom)

    def on_shutdown(self, _: Gtk.Application) -> None:
        self.video_handler.set_camera(None)

    def on_toggle_controls_clicked(self, btn: Gtk.Button) -> None:
        btn.set_icon_name(
            "controls-show-symbolic" if btn.get_active() else "controls-hide-symbolic"
        )
        self.toggle_presentation_mode(on=btn.get_active())

    def on_toggle_debug_position(self, button: Gtk.Button) -> None:
        debug_mode = button.get_active()
        self.video_handler.set_debug_mode(on=debug_mode)

    def on_picture_tick(self, widget: Gtk.Widget, _: Gdk.FrameClock) -> bool:
        """Tick callback on picture container.

        Used to update the webcam image on every application tick.

        Args:
            widget: Tick owner widget.
            idle: The frame clock for the widget.

        Returns:
            True if the tick callback should continue to be called.
        """
        self.draw_image(widget)
        return True

    def toggle_presentation_mode(self, on: bool) -> None:
        titlebar_height = self.win.get_titlebar().get_height()
        css_classes = self.win.get_css_classes()

        if on:
            css_classes.append("transparent")
            self.win.set_decorated(False)
            self.overlay.set_margin_top(titlebar_height)
        else:
            css_classes.remove("transparent")
            self.win.set_decorated(True)
            self.overlay.set_margin_top(0)

        self.win.set_css_classes(css_classes)

    def draw_image(self, widget: Gtk.Widget) -> None:
        """Draw webcam image on container widget.

        Args:
            widget: Tick owner widget.
        """
        tick_before = time.perf_counter()

        image = self.video_handler.get_processed_frame()

        # Compare an approx. image hash with the last one to avoid unnecessary updates:
        new_image_id = image[::100, ::100, 1].data.tobytes()
        if self.last_image_id != new_image_id:
            self.last_image_id = new_image_id

            self.left_button.set_sensitive(self.video_handler.can_move_left())
            self.right_button.set_sensitive(self.video_handler.can_move_right())
            self.up_button.set_sensitive(self.video_handler.can_move_up())
            self.down_button.set_sensitive(self.video_handler.can_move_down())
            self.zoom_out_button.set_sensitive(self.video_handler.can_zoom_out())
            self.zoom_in_button.set_sensitive(self.video_handler.can_zoom_in())

            height, width, channels = image.shape
            pixbuf = GdkPixbuf.Pixbuf.new_from_data(
                image.tobytes(),
                GdkPixbuf.Colorspace.RGB,
                True,
                8,
                width,
                height,
                width * channels,
            )
            texture = Gdk.Texture.new_for_pixbuf(pixbuf)
            widget.set_paintable(texture)

        if logger.getEffectiveLevel() <= logging.INFO:
            self.win.set_title(
                f"MyHumbleSelf - "
                f"FPS in/out: {mean(self.video_handler._camera.fps):.1f} "
                f"/ {mean(self.fps):.1f}"
            )

        tick_after = time.perf_counter()
        fps = 1 / (tick_after - tick_before)
        self.fps.append(fps)
        if len(self.fps) > self.fps_window:
            self.fps.pop(0)

    def get_system_info(self) -> str:
        xdg_session_type = os.environ.get("XDG_SESSION_TYPE", "").lower()
        has_wayland_display_env = bool(os.environ.get("WAYLAND_DISPLAY", ""))
        is_wayland = "wayland" in xdg_session_type or has_wayland_display_env

        linux_info = platform.freedesktop_os_release()

        text = f"Distro: {linux_info.get('NAME', 'Unknown')} "
        text += f"({linux_info.get('BUILD_ID', 'Unknown')})\n"
        text += f"DE: {self.get_desktop_environment()}\n"
        text += f"Wayland: {is_wayland}\n"
        text += f"Flatpak: {os.getenv('FLATPAK_ID') is not None}\n"
        text += f"Python: {platform.python_version()}\n"
        text += f"GTK: {Gtk.get_major_version()}.{Gtk.get_minor_version()}\n"

        logger.debug("System information:\n%s", text)
        return text

    def get_desktop_environment(self) -> str:  # noqa: PLR0911 # too many returns
        """Detect used desktop environment."""
        kde_full_session = os.environ.get("KDE_FULL_SESSION", "").lower()
        xdg_current_desktop = os.environ.get("XDG_CURRENT_DESKTOP", "").lower()
        desktop_session = os.environ.get("DESKTOP_SESSION", "").lower()
        gnome_desktop_session_id = os.environ.get("GNOME_DESKTOP_SESSION_ID", "")
        hyprland_instance_signature = os.environ.get("HYPRLAND_INSTANCE_SIGNATURE", "")

        if gnome_desktop_session_id == "this-is-deprecated":
            gnome_desktop_session_id = ""

        if gnome_desktop_session_id or "gnome" in xdg_current_desktop:
            return "Gnome"
        if kde_full_session or "kde-plasma" in desktop_session:
            return "KDE"
        if "sway" in xdg_current_desktop or "sway" in desktop_session:
            return "Sway"
        if "unity" in xdg_current_desktop:
            return "Unity"
        if hyprland_instance_signature:
            return "Hyprland"
        if "awesome" in xdg_current_desktop:
            return "Awesome"

        return "Unknown"


def _parse_args() -> argparse.Namespace:
    """Configure and process cli arguments.

    Returns:
        Parsed arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Enable info logging."
    )
    parser.add_argument(
        "-vv", "--very-verbose", action="store_true", help="Enable debug logging."
    )
    return parser.parse_args()


def main() -> None:
    args = _parse_args()

    if args.very_verbose:
        log_level = "DEBUG"
    elif args.verbose:
        log_level = "INFO"
    else:
        log_level = "WARNING"

    init_logger(log_level=log_level)

    app = MyHumbleSelf(application_id="com.github.dynobo.myhumbleself", args=args)
    app.run(None)


if __name__ == "__main__":
    main()



import threading
from collections.abc import Callable
from time import sleep

import pytest
from gi.repository import GLib

from myhumbleself import app


def get_value(func: Callable, key: str, result: dict):
    result[key] = func()


def do_actions(funcs: list[Callable]):
    delay = 200

    for idx, func in enumerate(funcs):
        GLib.timeout_add(idx * delay, func)

    sleep(1 + len(funcs) * delay / 1000)  # Wait for all actions to be executed


@pytest.fixture()
def _init_config(monkeypatch):
    monkeypatch.setattr(
        app.config.WritingConfigParser,
        "set_persistent",
        app.config.WritingConfigParser.set,
    )

    def load_parser():
        parser = app.config.WritingConfigParser(
            defaults={
                "zoom_factor": "1",
                "follow_face": "True",
                "offset_x": "0",
                "offset_y": "0",
                "shape": "01-circle.png",
                "last_active_camera": "0",
            },
        )
        parser.add_section("main")
        return parser

    monkeypatch.setattr(app.config, "load", load_parser)


@pytest.fixture()
def mhs_app(_init_config):
    mhs = app.MyHumbleSelf(application_id="com.github.dynobo.myhumbleself", args={})
    thread = threading.Thread(target=mhs.run)
    thread.start()

    sleep(1)  # Time to get up and running

    yield mhs

    GLib.idle_add(mhs.quit)
    thread.join()


@pytest.mark.gui()
def test_about_button(mhs_app):
    result = {}
    do_actions(
        [
            lambda: mhs_app.about_button.emit("clicked"),
            lambda: get_value(mhs_app.about.get_visible, "is_visible", result),
        ]
    )
    assert result["is_visible"]



import logging
from collections.abc import Callable

import cv2
import numpy as np

from myhumbleself import camera, face_detection, structures

logger = logging.getLogger(__name__)


def cache(func: Callable) -> Callable:
    """Custom approximating cache decorator for numpy arrays.

    This is intended to be used in VideoHandler._process_frame method only!

    The reason is, that often the GTK gui will call get_frame faster than the FPS of
    the camera. This will result in the same frame being processed multiple times.

    This decorator instead caches the processed frame and serves it, if the frame is
    still the same.

    To estimate, if the frame is the same, only one channel of every 100th pixel of
    the frame is taken into account. The number is a trade-off between performance and
    accuracy. Still, due to subtle movements of camera & person, and due to camera
    noise, the chance of a false positive seem low.

    Rough tests show a speedup of 2x (cache is hit ~once per frame), but it highly
    depends on camera FPS and system.

    (functools.lru_cache does not work with numpy arrays, as they are not hashable.
    hashing the whole image is too slow, anyway.)
    """
    cache.id = 0  # type: ignore [attr-defined]
    cache.content = None  # type: ignore [attr-defined]

    def inner(cls, ary: np.ndarray) -> np.ndarray:  # noqa: ANN001
        new_cache_id = ary[::100, ::100, 1].data.tobytes()  # type: ignore [attr-defined]
        if new_cache_id != cache.id:  # type: ignore [attr-defined]
            cache.id = new_cache_id  # type: ignore [attr-defined]
            cache.content = func(cls, ary)  # type: ignore [attr-defined]
        return cache.content  # type: ignore [attr-defined]

    return inner


class VideoHandler:
    def __init__(  # noqa:PLR0913
        self,
        cam_id: int,
        shape_png_buffer: bytes,
        zoom_factor: float,
        offset_x: int,
        offset_y: int,
        follow_face: bool,
    ) -> None:
        self._shape_mask = cv2.imdecode(
            np.frombuffer(shape_png_buffer, dtype=np.uint8), cv2.IMREAD_GRAYSCALE
        )
        # Face detection area. Cached here to allow the use case, where face detection
        # is only used once to get the face, but then disabled to avoid tracking during
        # presentation.
        self._face_area: structures.Rect | None = None
        # User adjusted area, with padding, offset and zoom. Cached to allow allow
        # detection, if we are already at the edge of the image, to disable buttons
        self._focus_area: structures.Rect | None = None

        self._camera = camera.Camera()
        self._face_detection = face_detection.FaceDetection()

        self.zoom_factor = zoom_factor
        self.offset_x = offset_x
        self.offset_y = offset_y

        self.follow_face = follow_face
        self.ZOOM_STEP = 0.1
        self.MOVE_STEP = 20
        self.MIN_ZOOM_FACTOR = 0.1
        self.debug_mode = False

        self.available_cameras = self._camera.available_cameras
        self.FALLBACK_CAM_ID = self._camera.FALLBACK_CAM_ID
        self.DEMO_CAM_ID = self._camera.DEMO_CAM_ID

        self._camera.start(cam_id)

    def _get_face_area_placeholder(self) -> structures.Rect:
        base_size = int(min(*self._frame_size_hw) / 1.6)
        rect = structures.Rect(
            top=(self._frame_size_hw[0] - base_size) // 2,
            left=(self._frame_size_hw[1] - base_size) // 2,
            width=base_size,
            height=base_size,
        )
        return rect

    @property
    def _frame_size_hw(self) -> tuple[int, int]:
        cam_image = self.available_cameras[self._camera.cam_id]
        return (cam_image.shape[0], cam_image.shape[1])

    def can_zoom_out(self) -> bool:
        if self._focus_area is None:
            return False
        return (
            self._focus_area.width < self._frame_size_hw[1]
            and self._focus_area.height < self._frame_size_hw[0]
        )

    def can_zoom_in(self) -> bool:
        return self.zoom_factor > self.MIN_ZOOM_FACTOR + self.ZOOM_STEP

    def can_move_left(self) -> bool:
        # Note: those methods are implemented for each direction individually, because
        # it is faster to call different methods than using one method with conditions.
        if self._focus_area is None:
            return False
        return self._focus_area.left > 0

    def can_move_right(self) -> bool:
        if self._focus_area is None:
            return False
        return self._focus_area.right < self._frame_size_hw[1]

    def can_move_up(self) -> bool:
        if self._focus_area is None:
            return False
        return self._focus_area.top > 0

    def can_move_down(self) -> bool:
        if self._focus_area is None:
            return False
        return self._focus_area.bottom < self._frame_size_hw[0]

    def set_camera(self, cam_id: int | None) -> None:
        self._camera.stop()
        if cam_id is not None:
            self._camera.start(cam_id)

    def set_shape(self, png_buffer: bytes) -> None:
        self._shape_mask = cv2.imdecode(
            np.frombuffer(png_buffer, dtype=np.uint8), cv2.IMREAD_GRAYSCALE
        )

    def set_debug_mode(self, on: bool) -> None:
        self.debug_mode = on
        self._face_detection.debug_mode = on

    def reset_view(self) -> None:
        self._face_area = None
        self.offset_x = 0
        self.offset_y = 0
        self.zoom_factor = 1.0

    def _draw_bbox(
        self,
        rect: structures.Rect,
        image: np.ndarray,
        color: tuple[int, int, int],
        label: str = "",
    ) -> None:
        border_width = 2
        cv2.rectangle(
            image,
            rect.left_top,
            (rect.right - border_width, rect.bottom - border_width),
            color,
            border_width,
        )
        cv2.putText(
            image,
            f"{label} {rect}",
            (rect.left + 10, rect.top + 30),
            cv2.FONT_HERSHEY_SIMPLEX,
            0.6,
            color,
            2,
        )

    def get_processed_frame(self) -> np.ndarray:
        frame = self._camera.get_frame()
        frame = self._process_frame(frame)
        return frame

    @cache
    def _process_frame(self, frame: np.ndarray) -> np.ndarray:
        """Process frame and return it with applied shape mask.

        Three different areas are calculated
        - face_area: Area supposed to contain face. Should be stabilized.
        - focus_area: User adjusted area, with padding, offset and zoom.
        - mask_area: Final area, restrained to image size. Should match aspect ratio of
          shape mask. At best case, this will be close to focus_area.

        Returns:
            Image ready to be displayed.
        """
        if self.follow_face:
            self._face_area = self._face_detection.get_face(frame)
        elif self._face_area is None:
            self._face_area = self._get_face_area_placeholder()

        self._focus_area = self._get_focus_area(face_area=self._face_area)
        mask_area = self._get_mask_area(
            focus_area=self._focus_area,
            image_size_hw=(frame.shape[0], frame.shape[1]),
            shape_size_hw=(self._shape_mask.shape[0], self._shape_mask.shape[1]),
        )

        if self.debug_mode:
            self._draw_bbox(
                rect=self._face_area,
                image=frame,
                color=(0, 255, 0),
                label="Face",
            )
            self._draw_bbox(
                rect=self._focus_area,
                image=frame,
                color=(255, 0, 0),
                label="Focus",
            )
            self._draw_bbox(
                rect=mask_area,
                image=frame,
                color=(0, 0, 255),
                label="Mask",
            )
            return cv2.cvtColor(frame, cv2.COLOR_BGR2RGBA)

        frame = self._crop_to_mask(image=frame, mask=mask_area)
        frame = self._apply_shape_mask(image=frame, shape_mask=self._shape_mask)
        return frame

    def _apply_shape_mask(
        self, image: np.ndarray, shape_mask: np.ndarray
    ) -> np.ndarray:
        image_height, image_width, _ = image.shape

        # Scale grayscale shape mask
        mask = cv2.resize(
            shape_mask, (image_width, image_height), interpolation=cv2.INTER_NEAREST
        )
        # ONHOLD: Converting to RGBA is very slow, but GdkPixbuf can't handle BGR
        # and OpenCV face detection doesn't work with RGB
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGBA)
        image[:, :, 3] = mask
        return image

    def _crop_to_mask(self, image: np.ndarray, mask: structures.Rect) -> np.ndarray:
        mask_height, mask_width = mask.height, mask.width
        mask_x, mask_y = mask.left, mask.top

        # Crop image to bounds
        cropped_image = image[
            mask_y : mask_y + mask_height, mask_x : mask_x + mask_width
        ]

        return cropped_image

    def _get_focus_area(self, face_area: structures.Rect) -> structures.Rect:
        focus_area = face_area.copy()
        focus_area.scale(self.zoom_factor)

        padding = int(max(face_area.width, face_area.height) / 1.5 * self.zoom_factor)
        focus_area.pad(padding=padding)

        focus_area.move_by(x=self.offset_x, y=self.offset_y)
        return focus_area

    def _get_mask_area(
        self,
        focus_area: structures.Rect,
        image_size_hw: tuple[int, int],
        shape_size_hw: tuple[int, int],
    ) -> structures.Rect:
        mask_area = focus_area.copy()
        shape_height, shape_width = shape_size_hw
        aspect_ratio = shape_width / shape_height

        # Adjust aspect ratio of mask area to match shape
        if mask_area.width / mask_area.height < aspect_ratio:
            new_width = int(mask_area.height * aspect_ratio)
            width_delta = new_width - mask_area.width
            mask_area.width = new_width
            mask_area.left = mask_area.left - width_delta // 2
        else:
            new_height = int(mask_area.width / aspect_ratio)
            height_delta = new_height - mask_area.height
            mask_area.height = new_height
            mask_area.top = mask_area.top - height_delta // 2

        image_height, image_width = image_size_hw
        mask_area.stay_within(width=image_width, height=image_height)
        return mask_area
